#include "RecognitionSummary.h"
#include <tracking/Tracker.h>
#include <gui/gui.h>
#include <tracking/Recognition.h>

using namespace track;

namespace gui {
    void RecognitionSummary::update(gui::DrawStructure& base) {
        auto & cache = GUI::instance()->cache();
        
        const float interface_scale = GUI_SETTINGS(gui_interface_scale);
        
        Font title_font(0.9 / interface_scale, Style::Bold, Align::Center);
        Font font(0.8 / interface_scale);
        Font side_font(0.8 / interface_scale, Align::Right);
        Font bottom_font(0.8 / interface_scale, Align::Center);
        
        auto manual_identities = FAST_SETTINGS(manual_identities);
        std::set<idx_t> sorted(manual_identities.begin(), manual_identities.end());
        for(auto id : manual_identities) {
            if(cache.individuals.find(id) == cache.individuals.end())
                sorted.erase(id);
        }

        size_t output_size = Tracker::recognition()->number_classes();
        obj.set_scale(base.scale().reciprocal().mul(interface_scale));

        const float margin = 5 / interface_scale,
        bar_width = 80 / (interface_scale * 1.25),
        title_height = Base::default_line_spacing(title_font) + margin;

        float sidebar_width = 0;
        for(auto id : sorted) {
            auto fish = cache.individuals.at(id);
            auto bds = Base::default_text_bounds(fish->identity().name(), &obj, side_font);
            sidebar_width = max(sidebar_width, bds.width);
        }
        sidebar_width += 3 * margin;

        obj.set_origin(Vec2(0.5));
        obj.set_bounds(Bounds(Vec2(Tracker::average().cols, Tracker::average().rows) * 0.5,
                              Size2(sidebar_width * 1.5, Base::default_line_spacing(font) + margin + title_height) + Size2(margin * 2) + bar_width * Size2(output_size, sorted.size())));
        obj.set_background(Black.alpha(150));

        if(!cache.recognition_updated) {
            obj.update([&] (Entangled& base) {
                std::vector<float> outputs;
                base.advance(new Text("recognition summary", Vec2(obj.width() * 0.5, margin + (title_height - margin) * 0.5), White, title_font));
                
                size_t counter = 0, j = 0;
                std::map<long_t, size_t> fdx_to_idx;
                std::map<size_t, long_t> idx_to_fdx;
                
                outputs.resize(output_size * sorted.size());
                
                for(auto id : sorted) {
                    auto fish = cache.individuals.at(id);
                    
                    float maxp = 0;
                    long_t idx = -1;
                    
                    for(auto && [fdx, p] : fish->average_recognition()) {
                        if(p > maxp) {
                            maxp = p;
                            idx = fdx;
                        }
                        
                        auto it = fdx_to_idx.find(fdx);
                        if(it == fdx_to_idx.end()) {
                            idx_to_fdx[counter] = fdx;
                            fdx_to_idx[fdx] = counter++;
                        }
                        auto idx = fdx_to_idx.at(fdx);
                        outputs.at(j * output_size + idx) = p;
                    }
                    
                    ++j;
                }
                
                auto image = std::make_unique<Image>(bar_width * sorted.size(), bar_width * output_size, 4);
                image->set_to(0);
                auto mat = image->get();
                
                for(size_t row = 0; row < sorted.size(); ++row) {
                    for (size_t j=0; j<output_size; ++j) {
                        auto p = idx_to_fdx.count(j) ? outputs.at(row * output_size + idx_to_fdx.at(j)) : 0;
                        
                        Color interp = Viridis::value(p).alpha(200);
                        cv::rectangle(mat,
                                      Vec2(j,row) * bar_width + Vec2(1),
                                      Vec2(j+1,row+1) * bar_width - Vec2(2),
                                      interp,
                                      -1);
                    }
                }
                
                auto pos = Vec2(margin + sidebar_width, margin + title_height);
                auto bounds = Bounds(pos, image->bounds().size());
                base.advance(new ExternalImage(std::move(image), pos));
                base.advance(new Rect(bounds, Transparent, White.alpha(200)));
                
                // draw vertical bar (active fish)
                pos = Vec2(margin) + Vec2(sidebar_width - 10 / interface_scale, bar_width * 0.5 - Base::default_line_spacing(font) * 0.5 + title_height);
                
                size_t row = 0;
                for(auto id : sorted) {
                    auto fish = cache.individuals.at(id);
                    base.advance(new Text(fish->identity().name(), pos + Vec2(0, bar_width) * row, White, side_font));
                    ++row;
                }
                
                // draw horizontal bar (matched fish from network)
                pos = Vec2(margin) + Vec2(sidebar_width + bar_width * 0.5, bounds.height + margin + Base::default_line_spacing(font) * 0.5 + title_height);
                for(size_t idx = 0; idx < output_size; ++idx) {
                    base.advance(new Text(Meta::toStr(idx), pos, White, bottom_font));
                    pos += Vec2(bar_width, 0);
                }
            });
            
            cache.recognition_updated = true;
        }
        
        base.wrap_object(obj);
    }
}
